@@ -33,6 +33,13 @@
 
 #ifdef HAVE_W32CE_SYSTEM
 # include "mkw32errmap.map.c"  /* Generated map_w32codes () */
+# ifndef TLS_OUT_OF_INDEXES
+#  define TLS_OUT_OF_INDEXES 0xFFFFFFFF
+# endif
+# ifndef __MINGW32CE__
+#  /* Replace the Mingw32CE provided abort function.  */
+#  define abort() do { TerminateProcess (GetCurrentProcess(), 8); } while (0)
+# endif
 #endif
 
 
@@ -40,7 +47,9 @@
 
 #if HAVE_W32_SYSTEM
 
-static int tls_index;  /* Index for the TLS functions.  */ 
+#include <windows.h>
+
+static int tls_index = TLS_OUT_OF_INDEXES;  /* Index for the TLS functions.  */ 
 
 static char *get_locale_dir (void);
 static void drop_locale_dir (char *locale_dir);
@@ -74,12 +83,30 @@ gpg_error_t
 gpg_err_init (void)
 {
 #ifdef HAVE_W32_SYSTEM
+# ifdef DLL_EXPORT
   /* We always have a constructor and thus this function is called
      automatically.  Due to the way the C init code of mingw works,
      the constructors are called before our DllMain function is
      called.  The problem with that is that the TLS has not been setup
      and w32-gettext.c requires TLS.  To solve this we do nothing here
      but call the actual init code from our DllMain.  */
+# else /*!DLL_EXPORT*/
+  /* Note that if the TLS is actually used, we can't release the TLS
+     as there is no way to know when a thread terminates (i.e. no
+     thread-specific-atexit).  You are really better off to use the
+     DLL! */
+  if (tls_index == TLS_OUT_OF_INDEXES)
+    {
+      tls_index = TlsAlloc ();
+      if (tls_index == TLS_OUT_OF_INDEXES)
+        {
+          /* No way to continue - commit suicide.  */
+          abort ();
+        }
+      _gpg_w32__init_gettext_module ();
+      real_init ();
+    }
+# endif /*!DLL_EXPORT*/
 #else
   real_init ();
 #endif
@@ -87,11 +114,41 @@ gpg_err_init (void)
 }
 
 
+/* Deinitialize libgpg-error.  This function is only used in special
+   circumstances.  No gpg-error function should be used after this
+   function has been called.  A value of 0 passed for MODE
+   deinitializes the entire libgpg-error, a value of 1 releases
+   resources allocated for the current thread and only that thread may
+   not anymore access libgpg-error after such a call.  Under Windows
+   this function may be called from the DllMain function of a DLL
+   which statically links to libgpg-error.  */
+void
+gpg_err_deinit (int mode)
+{
+#if defined (HAVE_W32_SYSTEM) && !defined(DLL_EXPORT)
+  struct tls_space_s *tls;
+  
+  tls = TlsGetValue (tls_index);
+  if (tls)
+    {
+      TlsSetValue (tls_index, NULL);
+      LocalFree (tls);
+    }
+
+  if (mode == 0)
+    {
+      TlsFree (tls_index);
+      tls_index = TLS_OUT_OF_INDEXES;
+    }
+#else
+  (void)mode;
+#endif
+}
+
+
 
 #ifdef HAVE_W32_SYSTEM
 
-#include <windows.h>
-
 /* Return a malloced string encoded in UTF-8 from the wide char input
    string STRING.  Caller must free this value.  Returns NULL on
    failure.  Caller may use GetLastError to get the actual error
@@ -185,6 +242,27 @@ get_locale_dir (void)
               p = strrchr (result, '\\');
               if (p)
                 *p = 0;
+              /* If we are installed below "bin" strip that part and
+                 use the top directory instead.
+
+                 Background: Under Windows we don't install GnuPG
+                 below bin/ but in the top directory with only share/,
+                 lib/, and etc/ below it.  One of the reasons is to
+                 keep the the length of the filenames at bay so not to
+                 increase the limited length of the PATH envvar.
+                 Another and more important reason, however, is that
+                 the very first GPG versions on W32 were installed
+                 into a flat directory structure and for best
+                 compatibility with these versions we didn't changed
+                 that later.  For WindowsCE we can right away install
+                 it under bin, though.  The hack with detection of the
+                 bin directory part allows us to eventually migrate to
+                 such a directory layout under plain Windows without
+                 the need to change libgpg-error.  */
+              p = strrchr (result, '\\');
+              if (p && !strcmp (p+1, "bin"))
+                *p = 0;
+              /* Append the static part.  */
               strcat (result, SLDIR);
             }
         }
@@ -259,6 +337,10 @@ _gpg_w32ce_strerror (int err)
 
   if (err == -1)
     err = _gpg_w32ce_get_errno ();
+
+  /* Note: On a German HTC Touch Pro2 device I also tried
+     LOCALE_USER_DEFAULT and LOCALE_SYSTEM_DEFAULT - both returned
+     English messages.  */
   if (FormatMessageW (FORMAT_MESSAGE_FROM_SYSTEM, NULL, err,
                       MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
                       tmpbuf, STRBUFFER_SIZE -1,
@@ -292,6 +374,7 @@ gpg_err_set_errno (int err)
 
 
 /* Entry point called by the DLL loader.  */
+#ifdef DLL_EXPORT
 int WINAPI
 DllMain (HINSTANCE hinst, DWORD reason, LPVOID reserved)
 {
@@ -304,6 +387,10 @@ DllMain (HINSTANCE hinst, DWORD reason, LPVOID reserved)
       tls_index = TlsAlloc ();
       if (tls_index == TLS_OUT_OF_INDEXES)
         return FALSE; 
+#ifndef _GPG_ERR_HAVE_CONSTRUCTOR
+      /* If we have not constructors (e.g. MSC) we call it here.  */
+      _gpg_w32__init_gettext_module ();
+#endif
       /* falltru.  */
     case DLL_THREAD_ATTACH:
       tls = LocalAlloc (LPTR, sizeof *tls);
@@ -336,7 +423,7 @@ DllMain (HINSTANCE hinst, DWORD reason, LPVOID reserved)
   
   return TRUE;
 }
-
+#endif /*DLL_EXPORT*/
 
 #else /*!HAVE_W32_SYSTEM*/
 
